"""
This module provides functions to easily access ragged-array datasets.
"""

from clouddrift import adapters
import os
import xarray as xr


def gdp1h() -> xr.Dataset:
    """Returns the NOAA Global Drifter Program (GDP) hourly dataset as an Xarray
    dataset.

    The data is accessed from a public AWS S3 bucket accessible at
    https://registry.opendata.aws/noaa-oar-hourly-gdp/. This dataset includes
    corrections and additional metadata since the original submission of the
    dataset to NCEI (accessible via https://doi.org/10.25921/x46c-3620). We
    recommend using this dataset over the one distributed via NCEI.

    Returns
    -------
    xarray.Dataset
        Hourly GDP dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import gdp1h
    >>> ds = gdp1h()
    >>> ds
    <xarray.Dataset>
    Dimensions:                (traj: 17324, obs: 165754333)
    Coordinates:
        ids                    (obs) int64 ...
        lat                    (obs) float32 ...
        lon                    (obs) float32 ...
        time                   (obs) datetime64[ns] ...
    Dimensions without coordinates: traj, obs
    Data variables: (12/55)
        BuoyTypeManufacturer   (traj) |S20 ...
        BuoyTypeSensorArray    (traj) |S20 ...
        CurrentProgram         (traj) float64 ...
        DeployingCountry       (traj) |S20 ...
        DeployingShip          (traj) |S20 ...
        DeploymentComments     (traj) |S20 ...
        ...                     ...
        sst1                   (obs) float64 ...
        sst2                   (obs) float64 ...
        typebuoy               (traj) |S10 ...
        typedeath              (traj) int8 ...
        ve                     (obs) float32 ...
        vn                     (obs) float32 ...
    Attributes: (12/16)
        Conventions:       CF-1.6
        acknowledgement:   Elipot, Shane; Sykulski, Adam; Lumpkin, Rick; Centurio...
        contributor_name:  NOAA Global Drifter Program
        contributor_role:  Data Acquisition Center
        date_created:      2022-12-09T06:02:29.684949
        doi:               10.25921/x46c-3620
        ...                ...
        processing_level:  Level 2 QC by GDP drifter DAC
        publisher_email:   aoml.dftr@noaa.gov
        publisher_name:    GDP Drifter DAC
        publisher_url:     https://www.aoml.noaa.gov/phod/gdp
        summary:           Global Drifter Program hourly data
        title:             Global Drifter Program hourly drifting buoy collection

    See Also
    --------
    :func:`gdp6h`
    """
    url = "https://noaa-oar-hourly-gdp-pds.s3.amazonaws.com/latest/gdp_v2.00.zarr"
    return xr.open_dataset(url, engine="zarr")


def gdp6h() -> xr.Dataset:
    """Returns the NOAA Global Drifter Program (GDP) 6-hourly dataset as an
    Xarray dataset.

    The data is accessed from a public HTTPS server at NOAA's Atlantic
    Oceanographic and Meteorological Laboratory (AOML) accessible at
    https://www.aoml.noaa.gov/phod/gdp/index.php.

    Returns
    -------
    xarray.Dataset
        6-hourly GDP dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import gdp6h
    >>> ds = gdp6h()
    >>> ds
    <xarray.Dataset>
    Dimensions:                (traj: 26843, obs: 44544647)
    Coordinates:
        ids                    (obs) int64 ...
        time                   (obs) datetime64[ns] ...
        lon                    (obs) float32 ...
        lat                    (obs) float32 ...
    Dimensions without coordinates: traj, obs
    Data variables: (12/44)
        ID                     (traj) int64 ...
        rowsize                (traj) int32 ...
        WMO                    (traj) int32 ...
        expno                  (traj) int32 ...
        deploy_date            (traj) datetime64[ns] ...
        deploy_lat             (traj) float32 ...
        ...                     ...
        vn                     (obs) float32 ...
        temp                   (obs) float32 ...
        err_lat                (obs) float32 ...
        err_lon                (obs) float32 ...
        err_temp               (obs) float32 ...
        drogue_status          (obs) bool ...
    Attributes: (12/16)
        title:             Global Drifter Program six-hourly drifting buoy collec...
        history:           Last update July 2022.  Metadata from dirall.dat and d...
        Conventions:       CF-1.6
        date_created:      2022-12-08T18:44:27.784441
        publisher_name:    GDP Drifter DAC
        publisher_email:   aoml.dftr@noaa.gov
        ...                ...
        contributor_name:  NOAA Global Drifter Program
        contributor_role:  Data Acquisition Center
        institution:       NOAA Atlantic Oceanographic and Meteorological Laboratory
        acknowledgement:   Lumpkin, Rick; Centurioni, Luca (2019). NOAA Global Dr...
        summary:           Global Drifter Program six-hourly data
        doi:               10.25921/7ntx-z961

    See Also
    --------
    :func:`gdp1h`
    """
    url = "https://www.aoml.noaa.gov/ftp/pub/phod/buoydata/gdp_jul22_ragged_6h.nc#mode=bytes"
    return xr.open_dataset(url)


def mosaic() -> xr.Dataset:
    """Returns the MOSAiC sea-ice drift dataset as an Xarray dataset.

    The function will first look for the ragged-array dataset on the local
    filesystem. If it is not found, the dataset will be downloaded using the
    corresponding adapter function and stored for later access.

    The upstream data is available at https://arcticdata.io/catalog/view/doi:10.18739/A2KP7TS83.

    Reference: Angela Bliss, Jennifer Hutchings, Philip Anderson, Philipp Anhaus,
    Hans Jakob Belter, Jørgen Berge, Vladimir Bessonov, Bin Cheng, Sylvia Cole,
    Dave Costa, Finlo Cottier, Christopher J Cox, Pedro R De La Torre, Dmitry V Divine,
    Gilbert Emzivat, Ying-Chih Fang, Steven Fons, Michael Gallagher, Maxime Geoffrey,
    Mats A Granskog, ... Guangyu Zuo. (2022). Sea ice drift tracks from the Distributed
    Network of autonomous buoys deployed during the Multidisciplinary drifting Observatory
    for the Study of Arctic Climate (MOSAiC) expedition 2019 - 2021. Arctic Data Center.
    doi:10.18739/A2KP7TS83.

    Returns
    -------
    xarray.Dataset
        MOSAiC sea-ice drift dataset as a ragged array

    Examples
    --------
    >>> from clouddrift.datasets import mosaic
    >>> ds = mosaic()
    >>> ds
    <xarray.Dataset>
    Dimensions:                     (obs: 1926226, traj: 216)
    Coordinates:
        time                        (obs) datetime64[ns] ...
        id                          (traj) object ...
    Dimensions without coordinates: obs, traj
    Data variables: (12/19)
        latitude                    (obs) float64 ...
        longitude                   (obs) float64 ...
        Deployment Leg              (traj) int64 ...
        DN Station ID               (traj) object ...
        IMEI                        (traj) object ...
        Deployment Date             (traj) datetime64[ns] ...
        ...                          ...
        Buoy Type                   (traj) object ...
        Manufacturer                (traj) object ...
        Model                       (traj) object ...
        PI                          (traj) object ...
        Data Authors                (traj) object ...
        count                       (traj) int64 ...
    """
    clouddrift_path = (
        os.path.expanduser("~/.clouddrift")
        if not os.getenv("CLOUDDRIFT_PATH")
        else os.getenv("CLOUDDRIFT_PATH")
    )
    mosaic_path = f"{clouddrift_path}/data/mosaic.nc"
    if not os.path.exists(mosaic_path):
        print(f"{mosaic_path} not found; download from upstream repository.")
        ds = adapters.mosaic.to_xarray()
        os.makedirs(os.path.dirname(mosaic_path), exist_ok=True)
        ds.to_netcdf(mosaic_path)
    else:
        ds = xr.open_dataset(mosaic_path)
    return ds



import unittest
from unittest import TestCase
import os
import xarray as xr
import numpy as np
from clouddrift import RaggedArray
import awkward as ak

NETCDF_ARCHIVE = "test_archive.nc"
PARQUET_ARCHIVE = "test_archive.parquet"

if __name__ == "__main__":
    unittest.main()


class raggedarray_tests(TestCase):
    def setUp(self):
        """
        Create ragged array and output netCDF and Parquet file
        """
        self.drifter_id = [1, 2, 3]
        self.count = [10, 8, 2]
        self.nb_obs = np.sum(self.count)
        self.nb_traj = len(self.drifter_id)
        self.attrs_global = {
            "title": "test trajectories",
            "history": "version xyz",
        }
        self.variables_coords = ["ids", "time", "lon", "lat"]

        # append xr.Dataset to a list
        list_ds = []
        for i in range(0, len(self.count)):
            xr_coords = {}
            for var in ["lon", "lat", "time"]:
                xr_coords[var] = (
                    ["obs"],
                    np.random.rand(self.count[i]),
                    {"long_name": f"variable {var}", "units": "-"},
                )
            xr_coords["ids"] = (
                ["obs"],
                np.ones(self.count[i], dtype="int") * self.drifter_id[i],
                {"long_name": f"variable ids", "units": "-"},
            )

            xr_data = {}
            xr_data["ID"] = (
                ["traj"],
                [self.drifter_id[i]],
                {"long_name": f"variable ID", "units": "-"},
            )
            xr_data["count"] = (
                ["traj"],
                [self.count[i]],
                {"long_name": f"variable count", "units": "-"},
            )
            xr_data["temp"] = (
                ["obs"],
                np.random.rand(self.count[i]),
                {"long_name": f"variable temp", "units": "-"},
            )

            list_ds.append(
                xr.Dataset(coords=xr_coords, data_vars=xr_data, attrs=self.attrs_global)
            )

        # create test ragged array
        self.ra = RaggedArray.from_files(
            [0, 1, 2],
            lambda i: list_ds[i],
            self.variables_coords,
            ["ID", "count"],
            ["temp"],
        )

        # output archive
        self.ra.to_netcdf(NETCDF_ARCHIVE)
        self.ra.to_parquet(PARQUET_ARCHIVE)

    def tearDown(self):
        """
        Clean up saved archives
        """
        os.remove(NETCDF_ARCHIVE)
        os.remove(PARQUET_ARCHIVE)

    def test_from_awkward(self):
        ra = RaggedArray.from_awkward(ak.from_parquet(PARQUET_ARCHIVE))
        self.compare_awkward_array(ra.to_awkward())

    def test_from_xarray(self):
        ra = RaggedArray.from_xarray(xr.open_dataset(NETCDF_ARCHIVE))
        self.compare_awkward_array(ra.to_awkward())

    def test_from_xarray_dim_names(self):
        ds = xr.open_dataset("test_archive.nc")
        ra = RaggedArray.from_xarray(
            ds.rename_dims({"traj": "t", "obs": "o"}), dim_traj="t", dim_obs="o"
        )
        self.compare_awkward_array(ra.to_awkward())

    def test_length_ragged_arrays(self):
        """
        Validate the size of the ragged array variables
        """
        for var in ["lon", "lat", "time", "ids"]:
            self.assertEqual(len(self.ra.coords[var]), self.nb_obs)
        self.assertEqual(len(self.ra.metadata["ID"]), self.nb_traj)
        self.assertEqual(len(self.ra.data["temp"]), self.nb_obs)

    def test_variable_attrs(self):
        """
        Validate the variable attributes are properly transferred to the ragged array object.
        Note: as part of this test `long_name` is variable but `units` are always "-"
        """
        for var in ["lon", "lat", "time"]:
            self.assertEqual(
                self.ra.attrs_variables[var]["long_name"],
                f"variable {var}",
            )
            self.assertEqual(self.ra.attrs_variables[var]["units"], "-")

        for var in ["ids", "ID", "temp"]:
            self.assertEqual(
                self.ra.attrs_variables[var]["long_name"], f"variable {var}"
            )
            self.assertEqual(self.ra.attrs_variables[var]["units"], "-")

    def test_global_attrs(self):
        """
        Validate the global attributes are properly transferred to the ragged array object
        """
        self.assertEqual(self.ra.attrs_global, self.attrs_global)

    def compare_awkward_array(self, ds):
        """
        Compare the returned Awkward Array after initializing from a netCDF/Parquet archive
        with the ragged array object
        """
        # dimensions
        self.assertEqual(len(self.ra.data["temp"]), len(ak.flatten(ds.obs["temp"])))
        self.assertEqual(len(self.ra.metadata["ID"]), len(ds.obs["temp"]))

        # coords
        for var in ["lon", "lat", "time", "ids"]:
            self.assertTrue(
                np.allclose(self.ra.coords[var], ak.flatten(ds.obs[var]).to_numpy())
            )
            self.assertTrue(
                self.ra.attrs_variables[var] == ds.obs[var].layout.parameters["attrs"]
            )

        # metadata and
        self.assertTrue(np.allclose(self.ra.metadata["ID"], ds["ID"].to_numpy()))
        self.assertTrue(
            self.ra.attrs_variables["ID"] == ds["ID"].layout.parameters["attrs"]
        )

        # data
        self.assertTrue(
            np.allclose(self.ra.data["temp"], ak.flatten(ds.obs["temp"]).to_numpy())
        )
        self.assertTrue(
            self.ra.attrs_variables["temp"] == ds.obs["temp"].layout.parameters["attrs"]
        )

    def test_netcdf_output(self):
        """
        Validate the netCDF output archive
        """
        ds = RaggedArray.from_netcdf(NETCDF_ARCHIVE)
        self.compare_awkward_array(ds.to_awkward())

    def test_parquet_output(self):
        """
        Validate the netCDF output archive
        """
        ds = RaggedArray.from_parquet(PARQUET_ARCHIVE)
        self.compare_awkward_array(ds.to_awkward())



"""
This module provides signal processing functions.
"""

import numpy as np
from typing import Optional, Tuple, Union
import xarray as xr
import pandas as pd
import warnings


def analytic_transform(
    x: Union[list, np.ndarray, xr.DataArray, pd.Series],
    boundary: Optional[str] = "mirror",
) -> np.ndarray:
    """Return the analytic part of a real-valued signal or of a complex-valued
    signal. To obtain the anti-analytic part of a complex-valued signal apply analytic_transform
    to the conjugate of the input. Analytic_transform removes the mean of the input signals.

    Parameters
    ----------
    x : np.ndarray
        Real- or complex-valued signal
    boundary : str, optional ["mirror", "zeros", "periodic"] optionally specifies the
    boundary condition to be imposed at the edges of the time series. Default is "mirror".

    Returns
    -------
    z : np.ndarray
        Analytic transform of the input signal.

    Examples
    --------

    To obtain the analytic part of a real-valued signal:
    >>> x = np.random.rand(99)
    >>> z = analytic_transform(x)

    To obtain the analytic and anti-analytic parts of a complex-valued signal:
    >>> z = np.random.rand(99)+1j*np.random.rand(99)
    >>> zp = analytic_transform(z)
    >>> zn = analytic_transform(np.conj(z))

    To specify that a periodic boundary condition should be used
    >>> x = np.random.rand(99)
    >>> z = analytic_transform(x,boundary="periodic")

    Raises
    ------
    ValueError
        If ``boundary not in ["mirror", "zeros", "periodic"]``.
    """
    # assume unidimensional input; add dimension option
    m0 = len(x)

    # remove mean
    x = x - np.mean(x)

    # apply boundary conditions
    if boundary == "mirror":
        x = np.concatenate((np.flip(x), x, np.flip(x)))
    elif boundary == "zeros":
        x = np.concatenate((np.zeros_like(x), x, np.zeros_like(x)))
    elif boundary == "periodic":
        x = np.concatenate((x, x, x))
    else:
        raise ValueError("boundary must be one of 'mirror', 'align', or 'zeros'.")

    if np.isrealobj(x):
        z = 2 * np.fft.fft(x)
    else:
        z = np.fft.fft(x)

    m = len(x)

    # zero negative frequencies
    if m % 2 == 0:
        z[int(m / 2 + 2) - 1 : int(m + 1) + 1] = 0
    else:
        z[int((m + 3) / 2) - 1 : int(m + 1) + 1] = 0

    # inverse Fourier transform
    z = np.fft.ifft(z)

    # return central part
    z = z[int(m0 + 1) - 1 : int(2 * m0 + 1) - 1]

    return z


def rotary_transform(
    u: Union[list, np.ndarray, xr.DataArray, pd.Series],
    v: Union[list, np.ndarray, xr.DataArray, pd.Series],
    boundary: Optional[str] = "mirror",
) -> Tuple[np.ndarray, np.ndarray]:
    """Return time-domain rotary components time series (zp,zn) from Cartesian components time series (u,v).
    Note that zp and zn are both analytic time series which implies that the complex-valued time series
    u+1j*v is recovered by zp+np.conj(zn). The mean of the original complex signal is split evenly between
    the two rotary components.

    If up is the analytic transform of u, and vp the analytic transform of v, then the counterclockwise and
    clockwise components are defined by zp = 0.5*(up+1j*vp), zp = 0.5*(up-1j*vp).
    See as an example Lilly and Olhede (2010), doi: 10.1109/TSP.2009.2031729.

    Parameters
    ----------
    u : np.ndarray
        Real-valued signal, first Cartesian component (zonal, east-west)
    v : np.ndarray
        Real-valued signal, second Cartesian component (meridional, north-south)
    boundary : str, optional ["mirror", "zeros", "periodic"] optionally specifies the
    boundary condition to be imposed at the edges of the time series for the underlying analytic
    transform. Default is "mirror".

    Returns
    -------
    zp : np.ndarray
        Time-domain complex-valued positive (counterclockwise) rotary component.
    zn : np.ndarray
        Time-domain complex-valued negative (clockwise) rotary component.

    Examples
    --------

    To obtain the rotary components of a real-valued signal:
    >>> zp, zn = rotary_transform(u,v)

    Raises
    ------
    ValueError
        If the input arrays do not have the same shape.

    See Also
    --------
    :func:`analytic_transform`
    """
    # to implement: input one complex argument instead of two real arguments
    # u and v arrays must have the same shape or list length.
    if type(u) == list and type(v) == list:
        if not len(u) == len(v):
            raise ValueError("u and v must have the same length.")
    else:
        if not u.shape == v.shape:
            raise ValueError("u and v must have the same shape.")

    muv = np.mean(u) + 1j * np.mean(v)

    if muv == xr.DataArray:
        muv = muv.to_numpy()

    up = analytic_transform(u, boundary=boundary)
    vp = analytic_transform(v, boundary=boundary)

    return 0.5 * (up + 1j * vp) + 0.5 * muv, 0.5 * (up - 1j * vp) + 0.5 * np.conj(muv)



from clouddrift.signal import (
    analytic_transform,
    rotary_transform,
)
import numpy as np
import unittest
import pandas as pd
import xarray as xr

if __name__ == "__main__":
    unittest.main()


def test_analytic_transform_real_odd(self):
    x = np.random.rand(99)
    z = analytic_transform(x)
    self.assertTrue(np.allclose(x - np.mean(x), np.real(z)))


def test_analytic_transform_real_even(self):
    x = np.random.rand(100)
    z = analytic_transform(x)
    self.assertTrue(np.allclose(x - np.mean(x), np.real(z)))


def test_analytic_transform_imag_odd(self):
    z = np.random.rand(99) + 1j * np.random.rand(99)
    zp = analytic_transform(z)
    zn = analytic_transform(np.conj(z))
    self.assertTrue(np.allclose(z - np.mean(z), zp + np.conj(zn)))


def test_analytic_transform_imag_even(self):
    z = np.random.rand(100) + 1j * np.random.rand(100)
    zp = analytic_transform(z)
    zn = analytic_transform(np.conj(z))
    self.assertTrue(np.allclose(z - np.mean(z), zp + np.conj(zn)))


def test_analytic_transform_boundary(self):
    x = np.random.rand(99)
    z1 = analytic_transform(x, boundary="mirror")
    z2 = analytic_transform(x, boundary="zeros")
    z3 = analytic_transform(x, boundary="periodic")
    self.assertTrue(np.allclose(x - np.mean(x), np.real(z1)))
    self.assertTrue(np.allclose(x - np.mean(x), np.real(z2)))
    self.assertTrue(np.allclose(x - np.mean(x), np.real(z3)))


def test_analytic_transform_list(self):
    x = list(np.random.rand(99))
    z = analytic_transform(x)
    self.assertTrue(np.allclose(x - np.mean(x), np.real(z)))


def test_analytic_transform_array(self):
    x = np.random.rand(99)
    z = analytic_transform(x)
    self.assertTrue(np.allclose(x - np.mean(x), np.real(z)))


def test_analytic_transform_pandas(self):
    x = pd.Series(data=np.random.rand(99))
    z = analytic_transform(x)
    self.assertTrue(np.allclose(x - np.mean(x), np.real(z)))


def test_analytic_transform_xarray(self):
    x = xr.DataArray(data=np.random.rand(99))
    z = analytic_transform(x)
    self.assertTrue(np.allclose(x - np.mean(x), np.real(z)))


def test_rotary_transform_array(self):
    u = np.random.rand(99)
    v = np.random.rand(99)
    zp, zn = rotary_transform(u, v)
    self.assertTrue(np.allclose(u + 1j * v, zp + np.conj(zn)))


def test_rotary_transform_list(self):
    u = list(np.random.rand(99))
    v = list(np.random.rand(99))
    zp, zn = rotary_transform(u, v)
    self.assertTrue(np.allclose(u + 1j * v, zp + np.conj(zn)))


def test_rotary_transform_pandas(self):
    u = pd.Series(data=np.random.rand(99))
    v = pd.Series(data=np.random.rand(99))
    zp, zn = rotary_transform(u, v)
    self.assertTrue(np.allclose(u + 1j * v, zp + np.conj(zn)))


def test_rotary_transform_xarray(self):
    u = xr.DataArray(data=np.random.rand(99))
    v = xr.DataArray(data=np.random.rand(99))
    zp, zn = rotary_transform(u, v)
    self.assertTrue(np.allclose(u + 1j * v, zp + np.conj(zn)))
